<div class="left-menu-wrapper">
    <div class="left-menu">
        <h3><a href="sql-programming-guide.html">Spark SQL Guide</a></h3>
        {% include nav-left.html nav=include.nav-sql %}
    </div>
</div>