/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.spark.ml.r

import org.apache.spark.mllib.stat.Statistics.kolmogorovSmirnovTest
import org.apache.spark.mllib.stat.test.KolmogorovSmirnovTestResult
import org.apache.spark.sql.{DataFrame, Row}

private[r] class KSTestWrapper private (
    val testResult: KolmogorovSmirnovTestResult,
    val distName: String,
    val distParams: Array[Double]) {

  lazy val pValue = testResult.pValue

  lazy val statistic = testResult.statistic

  lazy val nullHypothesis = testResult.nullHypothesis

  lazy val degreesOfFreedom = testResult.degreesOfFreedom

  def summary: String = testResult.toString
}

private[r] object KSTestWrapper {

  def test(
      data: DataFrame,
      featureName: String,
      distName: String,
      distParams: Array[Double]): KSTestWrapper = {

    val rddData = data.select(featureName).rdd.map {
      case Row(feature: Double) => feature
    }

    import org.apache.spark.util.ArrayImplicits._
    val ksTestResult = kolmogorovSmirnovTest(rddData, distName, distParams.toImmutableArraySeq : _*)

    new KSTestWrapper(ksTestResult, distName, distParams)
  }
}

