#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import tempfile
import unittest


class FailInSetUpClassTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        raise Exception("error")

    def test_something(self):
        self.assertEqual(True, True)


if __name__ == "__main__":
    from pyspark.testing import main

    # Do not keep XML report because the test is supposed to fail.
    # We do not want to confuse XML report consumers.
    with tempfile.TemporaryDirectory() as tmpdir:
        try:
            main(output=tmpdir)
        except SystemExit as e:
            assert e.code == 1, f"status code: {e.code}"
