/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.spark.sql.execution.streaming.operators.stateful.transformwithstate.timers

import org.apache.spark.sql.streaming.TimerValues

/**
 * Class that provides a concrete implementation for TimerValues used for fetching
 * processing time and event time (watermark).
 * @param currentProcessingTimeOpt - option to current processing time
 * @param currentWatermarkOpt - option to current watermark
 */
class TimerValuesImpl(
    currentProcessingTimeOpt: Option[Long],
    currentWatermarkOpt: Option[Long]) extends TimerValues {

  // Return available processing time or -1 otherwise
  override def getCurrentProcessingTimeInMs(): Long = currentProcessingTimeOpt.getOrElse(-1L)

  // Return available watermark or -1 otherwise
  override def getCurrentWatermarkInMs(): Long = currentWatermarkOpt.getOrElse(-1L)
}
