/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.spark.sql.execution.adaptive

import org.apache.spark.sql.catalyst.plans.QueryPlan

/**
 * Exception thrown when an invalid query plan is detected in AQE replanning,
 * in which case AQE will stop the current replanning process and keep using the latest valid plan.
 *
 * @param message The reason why the plan is considered invalid.
 * @param plan The invalid plan/sub-plan.
 */
case class InvalidAQEPlanException[QueryType <: QueryPlan[_]](message: String, plan: QueryType)
  extends Exception(message)
