/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.spark.api.r

import org.apache.spark.SparkFunSuite
import org.apache.spark.TestUtils.testCommandAvailable
import org.apache.spark.internal.config.R.SPARKR_COMMAND

class BaseRRunnerSuite extends SparkFunSuite {
  test("Retrieve R options from R command") {
    val rCommand = SPARKR_COMMAND.defaultValue.get
    assume(testCommandAvailable(rCommand))
    assert(BaseRRunner.getROptions(rCommand) === "--no-restore"
      || BaseRRunner.getROptions(rCommand) === "--vanilla")
  }

  test("Should return the default value if the R command does not exist") {
    assume(!testCommandAvailable("nonexistent"))
    assert(BaseRRunner.getROptions("nonexistent") === "--vanilla")
  }
}
