/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.spark.sql.execution.datasources.csv

import org.apache.spark.sql.Row
import org.apache.spark.sql.test.SharedSparkSession

class CSVParsingOptionsSuite extends SharedSparkSession {
  import testImplicits._

  test("SPARK-49955: null string value does not mean corrupted file") {
    val str = "abc"
    val stringDataset = Seq(str, null).toDS()
    val df = spark.read.csv(stringDataset)
    // `spark.read.csv(rdd)` removes all null values at the beginning.
    checkAnswer(df, Seq(Row("abc")))
    val df2 = spark.read.option("mode", "failfast").csv(stringDataset)
    checkAnswer(df2, Seq(Row("abc")))
  }
}
