/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.spark.sql

import java.util.concurrent.atomic.AtomicBoolean

import org.apache.spark.sql.execution.QueryExecution
import org.apache.spark.sql.util.QueryExecutionListener


class TestQueryExecutionListener extends QueryExecutionListener {
  override def onSuccess(funcName: String, qe: QueryExecution, durationNs: Long): Unit = {
    OnSuccessCall.isOnSuccessCalled.set(true)
  }

  override def onFailure(funcName: String, qe: QueryExecution, exception: Exception): Unit = { }
}

/**
 * This has a variable to check if `onSuccess` is actually called or not. Currently, this is for
 * the test case in PySpark. See SPARK-23942.
 */
object OnSuccessCall {
  val isOnSuccessCalled = new AtomicBoolean(false)

  def isCalled(): Boolean = isOnSuccessCalled.get()

  def clear(): Unit = isOnSuccessCalled.set(false)
}
